import sys
import pytest
from testfixtures import compare
import nbrmd
from .utils import list_all_notebooks


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.parametrize('nb_file', list_all_notebooks('.Rmd'))
def test_identity_write_read(nb_file):
    """
    Test that writing the notebook with ipynb, and read again, yields identity
    :param file:
    :return:
    """

    with open(nb_file) as fp:
        rmd = fp.read()

    nb = nbrmd.reads(rmd)
    rmd2 = nbrmd.writes(nb)

    compare(rmd, rmd2)


def test_two_blank_lines_as_cell_separator():
    rmd = """Some markdown
text


And a new cell
"""

    nb = nbrmd.reads(rmd)
    assert len(nb.cells) == 2
    assert nb.cells[0].cell_type == 'markdown'
    assert nb.cells[1].cell_type == 'markdown'
    assert nb.cells[0].source == 'Some markdown\ntext'
    assert nb.cells[1].source == 'And a new cell'



import os
import copy


def list_all_notebooks(ext, path=None):
    """
    :ext: desired extension
    :return: all notebooks in the directory of this script,
     with the desired extension
    """
    nb_path = os.path.dirname(os.path.abspath(__file__))
    if path:
        nb_path = os.path.join(nb_path, path)
    notebooks = []
    for nb_file in os.listdir(nb_path):
        _, nb_ext = os.path.splitext(nb_file)
        if nb_ext.lower() == ext.lower():
            notebooks.append(os.path.join(nb_path, nb_file))
    return notebooks


def remove_outputs(nb):
    nb = copy.deepcopy(nb)
    for cell in nb.cells:
        cell.outputs = None
        cell.execution_count = None
        if 'trusted' in cell['metadata']:
            del cell['metadata']['trusted']

    for k in ['nbformat', 'nbformat_minor']:
        if k in nb:
            del nb[k]

    return nb


def remove_outputs_and_header(nb):
    nb = remove_outputs(nb)
    nb['metadata'] = {}
    return nb



import os
import subprocess
import pytest
import nbrmd
from .utils import list_all_notebooks


@pytest.mark.skipif(isinstance(nbrmd.RMarkdownExporter, str),
                    reason=nbrmd.RMarkdownExporter)
@pytest.mark.parametrize('nb_file', list_all_notebooks('.ipynb'))
def test_nbconvert_and_read(nb_file):
    # Load notebook
    nb = nbrmd.readf(nb_file)

    # Export to Rmd using nbrmd package
    rmd1 = nbrmd.writes(nb)

    # Export to Rmd using nbconvert exporter
    rmd_exporter = nbrmd.RMarkdownExporter()
    (rmd2, resources) = rmd_exporter.from_notebook_node(nb)

    assert rmd1 == rmd2


pytest.importorskip('jupyter')


@pytest.mark.skipif(isinstance(nbrmd.RMarkdownExporter, str),
                    reason=nbrmd.RMarkdownExporter)
@pytest.mark.parametrize('nb_file', list_all_notebooks('.ipynb'))
def test_nbconvert_cmd_line(nb_file, tmpdir):
    rmd_file = str(tmpdir.join('notebook.Rmd'))

    subprocess.call(['jupyter', 'nbconvert', '--to', 'rmarkdown',
                     nb_file, '--output', rmd_file])

    assert os.path.isfile(rmd_file)

    nb = nbrmd.readf(nb_file)
    rmd1 = nbrmd.writes(nb)
    with open(rmd_file) as fp:
        rmd2 = fp.read()

    assert rmd1 == rmd2



import nbformat
import pytest
from testfixtures import compare
import nbrmd
from .utils import list_all_notebooks, remove_outputs


@pytest.mark.parametrize('nb_file', list_all_notebooks('.ipynb'))
def test_identity_source_write_read(nb_file):
    """
    Test that writing the notebook with rmd, and read again,
    is the same as removing outputs
    :param file:
    :return:
    """

    with open(nb_file) as fp:
        nb1 = nbformat.read(fp, as_version=4)

    py = nbrmd.writes(nb1, ext='.py')
    nb2 = nbrmd.reads(py, ext='.py')

    compare(remove_outputs(nb1), remove_outputs(nb2))



import pytest
from testfixtures import compare
import nbrmd
from .utils import list_all_notebooks


@pytest.mark.parametrize('py_file', list_all_notebooks('.py', '../nbrmd') +
                         list_all_notebooks('.py'))
def test_identity_source_write_read(py_file):
    with open(py_file) as fp:
        py = fp.read()

    nb = nbrmd.reads(py, ext='.py')
    py2 = nbrmd.writes(nb, ext='.py')

    compare(py, py2)



# -*- coding: utf-8 -*-

import nbrmd
from testfixtures import compare


def test_read_simple_file(pynb="""# ---
# title: Simple file
# ---

# Here we have some text
# And below we have some python code

def f(x):
    return x+1


def h(y):
    return y-1
"""):
    nb = nbrmd.reads(pynb, ext='.py')
    assert len(nb.cells) == 4
    assert nb.cells[0].cell_type == 'raw'
    assert nb.cells[0].source == '---\ntitle: Simple file\n---'
    assert nb.cells[1].cell_type == 'markdown'
    assert nb.cells[1].source == 'Here we have some text\n' \
                                 'And below we have some python code'
    assert nb.cells[2].cell_type == 'code'
    compare(nb.cells[2].source, '''def f(x):
    return x+1''')
    assert nb.cells[3].cell_type == 'code'
    compare(nb.cells[3].source, '''def h(y):
    return y-1''')

    pynb2 = nbrmd.writes(nb, ext='.py')
    compare(pynb, pynb2)


def test_read_less_simple_file(pynb="""# ---
# title: Less simple file
# ---

# Here we have some text
# And below we have some python code

# This is a comment about function f
def f(x):
    return x+1


# And a comment on h
def h(y):
    return y-1
"""):
    nb = nbrmd.reads(pynb, ext='.py')

    assert len(nb.cells) == 4
    assert nb.cells[0].cell_type == 'raw'
    assert nb.cells[0].source == '---\ntitle: Less simple file\n---'
    assert nb.cells[1].cell_type == 'markdown'
    assert nb.cells[1].source == 'Here we have some text\n' \
                                 'And below we have some python code'
    assert nb.cells[2].cell_type == 'code'
    compare(nb.cells[2].source,
            '# This is a comment about function f\n'
            'def f(x):\n'
            '    return x+1')
    assert nb.cells[3].cell_type == 'code'
    compare(nb.cells[3].source,
            '''# And a comment on h\ndef h(y):\n    return y-1''')

    pynb2 = nbrmd.writes(nb, ext='.py')
    compare(pynb, pynb2)


def test_read_non_pep8_file(pynb="""# ---
# title: Non-pep8 file
# ---

# This file is non-pep8 as the function below has
# two consecutive blank lines in its body

def f(x):


    return x+1
"""):
    nb = nbrmd.reads(pynb, ext='.py')

    assert len(nb.cells) == 3
    assert nb.cells[0].cell_type == 'raw'
    assert nb.cells[0].source == '---\ntitle: Non-pep8 file\n---'
    assert nb.cells[1].cell_type == 'markdown'
    assert nb.cells[1].source == 'This file is non-pep8 as ' \
                                 'the function below has\n' \
                                 'two consecutive blank lines ' \
                                 'in its body'
    assert nb.cells[2].cell_type == 'code'
    compare(nb.cells[2].source,
            'def f(x):\n\n\n'
            '    return x+1')

    pynb2 = nbrmd.writes(nb, ext='.py')
    compare(pynb, pynb2)


def test_read_multiline_comment(pynb="""'''This is a multiline
comment with "quotes", 'single quotes'
# and comments
and line breaks


and it ends here'''


1 + 1
"""):
    nb = nbrmd.reads(pynb, ext='.py')

    assert len(nb.cells) == 2
    assert nb.cells[0].cell_type == 'code'
    assert nb.cells[0].source == """'''This is a multiline
comment with "quotes", 'single quotes'
# and comments
and line breaks


and it ends here'''"""
    assert nb.cells[1].cell_type == 'code'
    assert nb.cells[1].source == '1 + 1'

    pynb2 = nbrmd.writes(nb, ext='.py')
    compare(pynb, pynb2)


def test_no_space_after_code(pynb=u"""# -*- coding: utf-8 -*-
# Markdown cell

def f(x):
    return x+1

# And a new cell, and non ascii contênt
"""):
    nb = nbrmd.reads(pynb, ext='.py')

    assert len(nb.cells) == 3
    assert nb.cells[0].cell_type == 'markdown'
    assert nb.cells[0].source == 'Markdown cell'
    assert nb.cells[1].cell_type == 'code'
    assert nb.cells[1].source == 'def f(x):\n    return x+1'
    assert nb.cells[2].cell_type == 'markdown'
    assert nb.cells[2].source == u'And a new cell, and non ascii contênt'

    pynb2 = nbrmd.writes(nb, ext='.py')
    compare(pynb, pynb2)


def test_read_write_script(pynb="""#!/usr/bin/env python
# coding=utf-8
print('Hello world')
"""):
    nb = nbrmd.reads(pynb, ext='.py')
    pynb2 = nbrmd.writes(nb, ext='.py')
    compare(pynb, pynb2)



from nbformat.v4.nbbase import new_code_cell, new_markdown_cell, new_notebook
from nbrmd.combine import combine_inputs_with_outputs


def test_combine():
    nb_source = new_notebook(
        cells=[new_markdown_cell('Markdown text'),
               new_code_cell('a=3'),
               new_code_cell('a+1'),
               new_code_cell('a+1'),
               new_markdown_cell('Markdown text'),
               new_code_cell('a+2')])

    nb_outputs = new_notebook(
        cells=[new_markdown_cell('Markdown text'),
               new_code_cell('a=3'),
               new_code_cell('a+1'),
               new_code_cell('a+2'),
               new_markdown_cell('Markdown text')])

    nb_outputs.cells[2].outputs = ['4']
    nb_outputs.cells[3].outputs = ['5']

    combine_inputs_with_outputs(nb_source, nb_outputs)

    assert nb_source.cells[2].outputs == ['4']
    assert nb_source.cells[3].outputs == []
    assert nb_source.cells[5].outputs == ['5']



"""Combine source and outputs from two notebooks
"""
from .cell_metadata import _IGNORE_METADATA


def combine_inputs_with_outputs(nb_source, nb_outputs):
    '''Copy outputs of the second notebook into
    the first one, for cells that have matching inputs'''

    remaining_output_cells = nb_outputs.cells
    for cell in nb_source.cells:
        if cell.cell_type != 'code':
            continue

        # Remove outputs to warranty that trust of returned
        # notebook is that of second notebook
        cell.execution_count = None
        cell.outputs = []

        # Fill outputs with that of second notebook
        for i, ocell in enumerate(remaining_output_cells):
            if ocell.cell_type == 'code' and cell.source == ocell.source:
                cell.execution_count = ocell.execution_count
                cell.outputs = ocell.outputs

                metadata = ocell.metadata
                cell.metadata.update({k: metadata[k] for k in metadata
                                      if metadata in _IGNORE_METADATA})
                remaining_output_cells = remaining_output_cells[(i + 1):]
                break



"""Parse header of text notebooks
"""

import re
import yaml
import nbformat
from nbformat.v4.nbbase import new_raw_cell

_HEADER_RE = re.compile(r"^---\s*$")
_BLANK_RE = re.compile(r"^\s*$")
_JUPYTER_RE = re.compile(r"^jupyter\s*:\s*$")
_LEFTSPACE_RE = re.compile(r"^\s")
_ENCODING_RE = re.compile('^[ \t\f]*#.*?coding[:=][ \t]*([-_.a-zA-Z0-9]+)')
_UTF8_HEADER = '# -*- coding: utf-8 -*-'


def _as_dict(metadata):
    if isinstance(metadata, nbformat.NotebookNode):
        return {k: _as_dict(metadata[k]) for k in metadata.keys()}
    return metadata


def encoding_and_executable(self, notebook):
    """
    Return encoding and executable lines for a notebook, if applicable
    :param self:
    :param notebook:
    :return:
    """
    lines = []
    metadata = notebook.get('metadata', {})

    if self.ext != '.Rmd' and 'executable' in metadata:
        lines.append('#!' + metadata['executable'])
        del metadata['executable']

    if 'encoding' in metadata:
        lines.append(metadata['encoding'])
        del metadata['encoding']
    elif self.ext != '.Rmd':
        for cell in notebook.cells:
            try:
                cell.source.encode('ascii')
            except (UnicodeEncodeError, UnicodeDecodeError):
                lines.append(_UTF8_HEADER)
                break

    return lines


def metadata_and_cell_to_header(self, nb):
    '''
    Return the text header corresponding to a notebook, and remove the
    first cell of the notebook if it contained the header
    '''

    header = []
    skipline = True

    if nb.cells:
        cell = nb.cells[0]
        if cell.cell_type == 'raw':
            lines = cell.source.strip('\n\t ').splitlines()
            if len(lines) >= 2 \
                    and _HEADER_RE.match(lines[0]) \
                    and _HEADER_RE.match(lines[-1]):
                header = lines[1:-1]
                skipline = not cell.metadata.get('noskipline', False)
                nb.cells = nb.cells[1:]

    metadata = _as_dict(nb.get('metadata', {}))

    if metadata:
        header.extend(yaml.safe_dump({'jupyter': metadata},
                                     default_flow_style=False).splitlines())

    if header:
        header = ['---'] + header + ['---']

    header = self.markdown_escape(header)

    if header and skipline:
        header += ['']

    return header


def header_to_metadata_and_cell(self, lines):
    '''
    Return the metadata, first cell of notebook, and next loc in text
    '''

    header = []
    jupyter = []
    injupyter = False
    ended = False
    metadata = {}
    start = 0
    i = -1

    for i, line in enumerate(lines):
        if i == 0 and line.startswith('#!'):
            metadata['executable'] = line[2:]
            start = i + 1
            continue
        if i == 0 or (i == 1 and not _ENCODING_RE.match(lines[0])):
            encoding = _ENCODING_RE.match(line)
            if encoding:
                if encoding.group(1) != 'utf-8':
                    raise ValueError('Encodings other than utf-8 '
                                     'are not supported')
                if line != _UTF8_HEADER:
                    metadata['encoding'] = line
                start = i + 1
                continue

        if not line.startswith(self.prefix):
            break

        line = self.markdown_unescape(line)

        if i == start:
            if _HEADER_RE.match(line):
                continue
            else:
                break

        if i > start and _HEADER_RE.match(line):
            ended = True
            break

        if _JUPYTER_RE.match(line):
            injupyter = True
        elif not _LEFTSPACE_RE.match(line):
            injupyter = False

        if injupyter:
            jupyter.append(line)
        else:
            header.append(line)

    if ended:
        if jupyter:
            metadata.update(yaml.load('\n'.join(jupyter))['jupyter'])

        skipline = True
        if len(lines) > i + 1:
            line = self.markdown_unescape(lines[i + 1])
            if not _BLANK_RE.match(line):
                skipline = False
            else:
                i = i + 1
        else:
            skipline = False

        if header:
            cell = new_raw_cell(source='\n'.join(['---'] + header + ['---']),
                                metadata={} if skipline else
                                {'noskipline': True})
        else:
            cell = None

        return metadata, cell, i + 1

    return metadata, None, start



import pytest
import nbrmd


def test_read_wrong_ext(tmpdir, nb_file='notebook.ext'):
    nb_file = tmpdir.join('nb_file')
    nb_file.write('{}')
    with pytest.raises(TypeError):
        nbrmd.readf(str(nb_file))


def test_write_wrong_ext(tmpdir, nb_file='notebook.ext'):
    nb_file = str(tmpdir.join('nb_file'))
    with pytest.raises(TypeError):
        nbrmd.writef(dict(), nb_file)


def test_readme():
    nbrmd.nbrmd.readme()



# # Specifications for Jupyter notebooks as python scripts

# ## Markdown (and raw) cells

# Markdown cells are escaped with a single quote. Two consecutive
# cells are separated with a blank line. Raw cells are not
# distinguished from markdown.

# ## Code cells

# Code cells are separated by one blank line from markdown cells.
# If a code cells follows a comment, then that comment belong to the
# code cell.

# For instance, this is a code cell that starts with a
# code comment, split on multiple lines
1 + 2

# Code cells are terminated with either
# - end of file
# - two blank lines if followed by an other code cell
# - one blank line if followed by a markdown cell

# Code cells can have blank lines, but no two consecutive blank lines (that's
# a cell break!). Below we have a cell with multiple instructions:

a = 3

a + 1

# ## Metadata in code cells

# In case a code cell has metadata information, it
# is represented in json format, escaped with '#+' or '# +'

# + {"scrolled": true}
a + 2



import sys
import pytest
import nbrmd
from testfixtures import compare

ACTIVE_ALL = {'.py': """# + {"active": "ipynb,py,R,Rmd"}
# This cell is active in all extensions
""",
              '.Rmd': """```{python active="ipynb,py,R,Rmd"}
# This cell is active in all extensions
```
""",
              '.R': """#+ active="ipynb,py,R,Rmd"
# This cell is active in all extensions
""",
              '.ipynb': {'cell_type': 'code',
                         'source': '# This cell is active in all extensions',
                         'metadata': {'active': 'ipynb,py,R,Rmd'},
                         'execution_count': None,
                         'outputs': []}}


@pytest.mark.parametrize('ext', ['.Rmd', '.py', '.R'])
def test_active_all(ext):
    nb = nbrmd.reads(ACTIVE_ALL[ext], ext=ext)
    assert len(nb.cells) == 1
    compare(nb.cells[0], ACTIVE_ALL['.ipynb'])
    compare(ACTIVE_ALL[ext], nbrmd.writes(nb, ext=ext))


ACTIVE_PY_IPYNB = {'.py': """# + {"active": "ipynb,py"}
# This cell is active in py and ipynb extensions
""",
                   '.Rmd': """```{python active="ipynb,py", eval=FALSE}
# This cell is active in py and ipynb extensions
```
""",
                   '.R': """#' ```{python active="ipynb,py", eval=FALSE}
#' # This cell is active in py and ipynb extensions
#' ```
""",
                   '.ipynb': {'cell_type': 'code',
                              'source': '# This cell is active in py and '
                                        'ipynb extensions',
                              'metadata': {'active': 'ipynb,py'},
                              'execution_count': None,
                              'outputs': []}}


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.parametrize('ext', ['.Rmd', '.py'])  # TODO: add R
def test_active_py_ipynb(ext):
    nb = nbrmd.reads(ACTIVE_PY_IPYNB[ext], ext=ext)
    assert len(nb.cells) == 1
    compare(nb.cells[0], ACTIVE_PY_IPYNB['.ipynb'])
    compare(ACTIVE_PY_IPYNB[ext], nbrmd.writes(nb, ext=ext))


ACTIVE_RMD = {'.py': """# # + {"active": "Rmd"}
# # This cell is active in Rmd only
""",
              '.Rmd': """```{python active="Rmd"}
# This cell is active in Rmd only
```
""",
              '.R': """#' ```{python active="Rmd", eval=FALSE}
#' # This cell is active in Rmd only
#' ```
""",
              '.ipynb': {'cell_type': 'raw',
                         'source': '# This cell is active in Rmd only',
                         'metadata': {'active': 'Rmd'}}}


@pytest.mark.parametrize('ext', ['.Rmd'])  # TODO: add R and py
def test_active_rmd(ext):
    nb = nbrmd.reads(ACTIVE_RMD[ext], ext=ext)
    assert len(nb.cells) == 1
    compare(nb.cells[0], ACTIVE_RMD['.ipynb'])
    compare(ACTIVE_RMD[ext], nbrmd.writes(nb, ext=ext))


ACTIVE_NOT_INCLUDE_RMD = {'.py': """# # + {"hide_output": true, "active": "Rmd"}
# # This cell is active in Rmd only
""",
                          '.Rmd': """```{python include=FALSE, active="Rmd"}
# This cell is active in Rmd only
```
""",
                          '.R': """#' ```{python include=FALSE, active="Rmd", eval=FALSE}
#' # This cell is active in Rmd only
#' ```
""",
                          '.ipynb':
                              {'cell_type': 'raw',
                               'source': '# This cell is active in Rmd only',
                               'metadata': {'active': 'Rmd',
                                            'hide_output': True}}}


@pytest.mark.parametrize('ext', ['.Rmd'])  # TODO: add R and py
def test_active_not_include_rmd(ext):
    nb = nbrmd.reads(ACTIVE_NOT_INCLUDE_RMD[ext], ext=ext)
    assert len(nb.cells) == 1
    compare(nb.cells[0], ACTIVE_NOT_INCLUDE_RMD['.ipynb'])
    compare(ACTIVE_NOT_INCLUDE_RMD[ext], nbrmd.writes(nb, ext=ext))



"""Determine notebook or cell language"""

import re

_JUPYTER_LANGUAGES = ['R', 'bash', 'sh', 'python', 'python2', 'python3',
                      'javascript', 'js', 'perl']
_JUPYTER_LANGUAGES_RE = [re.compile(r"^%%{}\s*".format(lang))
                         for lang in _JUPYTER_LANGUAGES]


def is_code(cell):
    """Is the current code a code cell?"""
    if cell.cell_type == 'code':
        return True
    if cell.cell_type == 'raw' and 'active' in cell.metadata:
        return True
    return False


def get_default_language(nb):
    """Return the default language of a notebook, and remove metadata
    'main_language' if that information is clear from notebook
    contents"""
    metadata = nb.metadata

    default_language = (metadata.get('main_language') or
                        metadata.get('language_info', {})
                        .get('name', 'python').lower())

    if 'main_language' in metadata:
        # is 'main language' redundant with kernel info?
        if metadata['main_language'] is \
                metadata.get('language_info', {}).get('name'):
            del metadata['main_language']
        # is 'main language' redundant with cell language?
        elif metadata.get('language_info', {}).get('name') is None:
            languages = dict(python=0.5)
            for cell in nb.cells:
                if cell.cell_type == 'code':
                    source = cell.source.splitlines()
                    language = default_language
                    if source:
                        for lang, pattern in zip(_JUPYTER_LANGUAGES,
                                                 _JUPYTER_LANGUAGES_RE):
                            if pattern.match(source[0]):
                                language = lang

                    languages[language] = 1 + languages.get(
                        language, 0.0)

            cell_main_language = max(languages, key=languages.get)
            if metadata['main_language'] == cell_main_language:
                del metadata['main_language']

    return default_language


def find_main_language(metadata, cells):
    """
    Main language for the given collection of cells
    :param metadata:
    :param cells:
    :return:
    """
    main_language = (metadata.get('main_language') or
                     metadata.get('language_info', {}).get('name'))
    if main_language is None:
        languages = dict(python=0.5)
        for cell in cells:
            if cell.cell_type == 'code':
                language = cell['metadata']['language']
                languages[language] = languages.get(language, 0.0) + 1

        main_language = max(languages, key=languages.get)

        # save main language when not given by kernel
        if main_language is not \
                metadata.get('language_info', {}).get('name'):
            metadata['main_language'] = main_language

    # Remove 'language' meta data and add a magic if not main language
    for cell in cells:
        if is_code(cell):
            language = cell['metadata']['language']
            del cell['metadata']['language']
            if language != main_language and \
                    language in _JUPYTER_LANGUAGES:
                cell['source'] = u'%%{}\n'.format(language) + cell['source']


def cell_language(source):
    """
    Return cell language
    :param source:
    :return:
    """
    if source:
        for lang, pattern in zip(_JUPYTER_LANGUAGES, _JUPYTER_LANGUAGES_RE):
            if pattern.match(source[0]):
                source.pop(0)
                return lang

    return None



"""Command line conversion tool `nbrmd`
"""

import os
import argparse
from nbformat import writes as ipynb_writes
from nbformat.reader import NotJSONError
from nbrmd import readf, writef
from nbrmd import writes as rmd_writes
from .combine import combine_inputs_with_outputs


def convert(nb_files, in_place=True, combine=True):
    """
    Export R markdown notebooks, or Jupyter notebooks, to the opposite format
    :param nb_files: one or more notebooks
    :param in_place: should result of conversion be stored in file
    with opposite extension?
    :param combine: should the current outputs of .ipynb file be preserved,
    when a cell with corresponding input is found in .Rmd?
    :return:
    """
    for nb_file in nb_files:
        file, ext = os.path.splitext(nb_file)
        if ext not in ['.ipynb', '.Rmd']:
            raise TypeError(
                'File {} is neither a Jupyter (.ipynb) nor a '
                'R Markdown (.Rmd) notebook'.format(nb_file))

        nb = readf(nb_file)

        if in_place:
            if ext == '.ipynb':
                nb_dest = file + '.Rmd'
                print('Jupyter notebook {} being converted to '
                      'R Markdown {}'.format(nb_file, nb_dest))
            else:
                msg = ''
                nb_dest = file + '.ipynb'
                if combine and os.path.isfile(nb_dest):
                    try:
                        nb_outputs = readf(nb_dest)
                        combine_inputs_with_outputs(nb, nb_outputs)
                        msg = '(outputs were preserved)'
                    except (IOError, NotJSONError) as err:
                        msg = '(outputs were not preserved: {})'.format(err)
                print('R Markdown {} being converted to '
                      'Jupyter notebook {} {}'
                      .format(nb_file, nb_dest, msg))
            writef(nb, nb_dest)
        else:
            if ext == '.ipynb':
                print(rmd_writes(nb))
            else:
                print(ipynb_writes(nb))


def cli(args=None):
    """
    Command line parser
    :param args:
    :return:
    """
    parser = argparse.ArgumentParser(description='Jupyter notebook '
                                                 'from/to R Markdown')
    parser.add_argument('notebooks',
                        help='One or more .ipynb or .Rmd notebook(s) '
                             'to be converted to the alternate form',
                        nargs='+')
    parser.add_argument('-i', '--in-place', action='store_true',
                        help='Store the result of conversion '
                             'to file with opposite extension')
    parser.add_argument('-p', '--preserve_outputs', action='store_true',
                        help='Preserve outputs of .ipynb '
                             'notebook when file exists and inputs match')
    return parser.parse_args(args)


def main():
    """
    Entry point for the nbrmd script
    :return:
    """
    args = cli()
    convert(args.notebooks, args.in_place, args.preserve_outputs)



"""
R markdown exporter for nbconvert
"""

from traitlets import default
from nbconvert.exporters import Exporter
import nbrmd


class RMarkdownExporter(Exporter):
    """
    Exports to a R markdown document (.Rmd)
    """

    @default('file_extension')
    def _file_extension_default(self):
        return '.Rmd'

    def from_notebook_node(self, nb, resources=None, **kw):
        resources = resources or {}
        resources['output_extension'] = self.file_extension
        return nbrmd.writes(nb), resources



import os
from shutil import copyfile
import pytest
import nbrmd
from nbrmd.cli import convert, cli
from .utils import list_all_notebooks, remove_outputs


@pytest.mark.parametrize('nb_file',
                         list_all_notebooks('.ipynb') +
                         list_all_notebooks('.Rmd'))
def test_cli_single_file(nb_file):
    assert cli([nb_file]).notebooks == [nb_file]


@pytest.mark.parametrize('nb_files', [list_all_notebooks('.ipynb') +
                                      list_all_notebooks('.Rmd')])
def test_cli_multiple_files(nb_files):
    assert cli(nb_files).notebooks == nb_files


@pytest.mark.parametrize('nb_file',
                         list_all_notebooks('.ipynb') +
                         list_all_notebooks('.Rmd'))
def test_convert_single_file_in_place(nb_file, tmpdir):
    nb_org = str(tmpdir.join(os.path.basename(nb_file)))
    base, ext = os.path.splitext(nb_org)
    nb_other = base + '.ipynb' if ext == '.Rmd' else base + '.Rmd'

    copyfile(nb_file, nb_org)
    convert([nb_org])

    nb1 = nbrmd.readf(nb_org)
    nb2 = nbrmd.readf(nb_other)

    remove_outputs(nb1) == remove_outputs(nb2)


@pytest.mark.parametrize('nb_file',
                         list_all_notebooks('.ipynb') +
                         list_all_notebooks('.Rmd'))
def test_convert_single_file(nb_file, capsys):
    convert([nb_file], False)

    out, err = capsys.readouterr()
    assert out != ''
    assert err == ''


@pytest.mark.parametrize('nb_files',
                         [list_all_notebooks('.ipynb') +
                          list_all_notebooks('.Rmd')])
def test_convert_multiple_file(nb_files, tmpdir):
    nb_orgs = []
    nb_others = []

    for nb_file in nb_files:
        nb_org = str(tmpdir.join(os.path.basename(nb_file)))
        base, ext = os.path.splitext(nb_org)
        nb_other = base + '.ipynb' if ext == '.Rmd' else base + '.Rmd'
        copyfile(nb_file, nb_org)
        nb_orgs.append(nb_org)
        nb_others.append(nb_other)

    convert(nb_orgs)

    for nb_org, nb_other in zip(nb_orgs, nb_others):
        nb1 = nbrmd.readf(nb_org)
        nb2 = nbrmd.readf(nb_other)
        remove_outputs(nb1) == remove_outputs(nb2)


def test_error_not_notebook(nb_file='notebook.ext'):
    with pytest.raises(TypeError):
        convert([nb_file])



import os
import pytest
from nbformat.v4.nbbase import new_notebook
from nbformat.validator import NotebookValidationError
from tornado.web import HTTPError
import nbrmd
from nbrmd.contentsmanager import RmdFileContentsManager
from .utils import list_all_notebooks, remove_outputs, \
    remove_outputs_and_header


@pytest.mark.parametrize('nb_file', list_all_notebooks('.ipynb'))
def test_rmd_is_ok(nb_file, tmpdir):
    nb = nbrmd.readf(nb_file)
    tmp_ipynb = 'notebook.ipynb'
    tmp_rmd = 'notebook.Rmd'

    nb.metadata['nbrmd_formats'] = ['.Rmd']

    cm = RmdFileContentsManager()
    cm.root_dir = str(tmpdir)

    cm.save(
        model=dict(type='notebook', content=nb),
        path=tmp_ipynb)

    nb2 = nbrmd.readf(str(tmpdir.join(tmp_rmd)))

    assert remove_outputs(nb) == remove_outputs(nb2)


@pytest.mark.parametrize('nb_file', list_all_notebooks('.Rmd'))
def test_ipynb_is_ok(nb_file, tmpdir):
    nb = nbrmd.readf(nb_file)
    tmp_ipynb = 'notebook.ipynb'
    tmp_rmd = 'notebook.Rmd'

    cm = RmdFileContentsManager()
    cm.root_dir = str(tmpdir)

    cm.save(
        model=dict(type='notebook', content=nb),
        path=tmp_rmd)

    nb2 = nbrmd.readf(str(tmpdir.join(tmp_ipynb)))

    assert remove_outputs(nb) == remove_outputs(nb2)


@pytest.mark.parametrize('nb_file', list_all_notebooks('.ipynb'))
def test_all_files_created(nb_file, tmpdir):
    nb = nbrmd.readf(nb_file)
    tmp_ipynb = 'notebook.ipynb'
    tmp_rmd = 'notebook.Rmd'
    tmp_py = 'notebook.py'
    nb.metadata['nbrmd_formats'] = ['.ipynb', '.Rmd', '.py']

    cm = RmdFileContentsManager()
    cm.root_dir = str(tmpdir)

    cm.save(
        model=dict(type='notebook', content=nb),
        path=tmp_ipynb)

    nb2 = nbrmd.readf(str(tmpdir.join(tmp_py)))
    assert remove_outputs_and_header(nb) == remove_outputs_and_header(nb2)

    nb3 = nbrmd.readf(str(tmpdir.join(tmp_rmd)))
    assert remove_outputs(nb) == remove_outputs(nb3)


def test_no_files_created_on_no_format(tmpdir):
    tmp_ipynb = 'notebook.ipynb'
    tmp_rmd = 'notebook.Rmd'
    tmp_py = 'notebook.py'

    cm = RmdFileContentsManager()
    cm.root_dir = str(tmpdir)
    cm.default_nbrmd_formats = ''

    cm.save(
        model=dict(type='notebook',
                   content=new_notebook(nbformat=4,
                                        metadata=dict())),
        path=tmp_ipynb)

    assert not os.path.isfile(str(tmpdir.join(tmp_py)))
    assert not os.path.isfile(str(tmpdir.join(tmp_rmd)))


def test_raise_on_wrong_format(tmpdir):
    tmp_ipynb = str(tmpdir.join('notebook.ipynb'))

    cm = RmdFileContentsManager()
    cm.root_dir = str(tmpdir)

    with pytest.raises(HTTPError):
        cm.save(
            model=dict(type='notebook',
                       content=new_notebook(nbformat=4,
                                            metadata=dict(
                                                nbrmd_formats=['.doc']))),
            path=tmp_ipynb)


def test_no_rmd_on_not_notebook(tmpdir):
    tmp_ipynb = 'notebook.ipynb'
    tmp_rmd = 'notebook.Rmd'

    cm = RmdFileContentsManager()
    cm.root_dir = str(tmpdir)
    cm.default_nbrmd_formats = '.Rmd'

    with pytest.raises(HTTPError):
        cm.save(model=dict(type='not notebook',
                           content=new_notebook()),
                path=tmp_ipynb)
    assert not os.path.isfile(str(tmpdir.join(tmp_rmd)))


def test_no_rmd_on_not_v4(tmpdir):
    tmp_ipynb = 'notebook.ipynb'
    tmp_rmd = 'notebook.Rmd'

    cm = RmdFileContentsManager()
    cm.root_dir = str(tmpdir)
    cm.default_nbrmd_formats = '.Rmd'

    with pytest.raises(NotebookValidationError):
        cm.save(model=dict(type='notebook',
                           content=new_notebook(nbformat=3)),
                path=tmp_rmd)

    assert not os.path.isfile(str(tmpdir.join(tmp_ipynb)))



import re
from testfixtures import compare
import nbrmd


def test_read_mostly_py_rmd_file(rmd="""# ---
# title: Simple file
# ---

```{python, echo=TRUE}
x = np.arange(0, 2*math.pi, eps)
```

```{python, echo=TRUE}
x = np.arange(0,1,eps)
y = np.abs(x)-.5
```

```{r}
ls()
```

```{r, results='asis'}
cat(stringi::stri_rand_lipsum(3), sep='\n\n')
```
"""):
    nb = nbrmd.reads(rmd, ext='.Rmd')
    assert nb.metadata == {'main_language': 'python'}
    assert nb.cells == [{'cell_type': 'markdown',
                         'source': '# ---\n# title: Simple file\n# ---',
                         'metadata': {}},
                        {'cell_type': 'code',
                         'metadata': {'hide_input': False},
                         'execution_count': None,
                         'source': 'x = np.arange(0, 2*math.pi, eps)',
                         'outputs': []},
                        {'cell_type': 'code',
                         'metadata': {'hide_input': False},
                         'execution_count': None,
                         'source': 'x = np.arange(0,1,eps)\ny = np.abs(x)-.5',
                         'outputs': []},
                        {'cell_type': 'code',
                         'metadata': {},
                         'execution_count': None,
                         'source': '%%R\nls()',
                         'outputs': []},
                        {'cell_type': 'code',
                         'metadata': {'results': "'asis'"},
                         'execution_count': None,
                         'source': "%%R\ncat(stringi::"
                                   "stri_rand_lipsum(3), sep='\n\n')",
                         'outputs': []}]

    rmd2 = nbrmd.writes(nb, ext='.Rmd')
    rmd2 = re.sub(r'```{r ', '```{r, ', rmd2)
    rmd2 = re.sub(r'```{python ', '```{python, ', rmd2)
    compare(rmd, rmd2)



import os
import shutil
import pytest
import nbrmd


@pytest.fixture()
def readme():
    return os.path.join(os.path.dirname(os.path.abspath(__file__)),
                        '..', 'README.md')


def test_open_readme(readme, tmpdir):
    rmd_file = str(tmpdir.join('README.Rmd'))
    shutil.copyfile(readme, rmd_file)
    nb = nbrmd.readf(rmd_file)



"""
Convert between R markdown chunk options and jupyter cell metadata.

metadata.hide_input and metadata.hide_output are documented here:
http://jupyter-contrib-nbextensions.readthedocs.io/en/latest/nbextensions/runtools/readme.html

TODO: Update this if a standard gets defined at
https://github.com/jupyter/notebook/issues/3700
"""

import ast
import json
import re
from copy import copy

_BOOLEAN_OPTIONS_DICTIONARY = [('hide_input', 'echo', True),
                               ('hide_output', 'include', True)]
_IGNORE_METADATA = ['collapsed', 'autoscroll', 'deletable',
                    'format', 'trusted']


def _r_logical_values(pybool):
    return 'TRUE' if pybool else 'FALSE'


class RLogicalValueError(Exception):
    """Incorrect value for R boolean"""
    pass


class RMarkdownOptionParsingError(Exception):
    """Error when parsing Rmd cell options"""
    pass


def _py_logical_values(rbool):
    if rbool in ['TRUE', 'T']:
        return True
    if rbool in ['FALSE', 'F']:
        return False
    raise RLogicalValueError


def metadata_to_rmd_options(language, metadata):
    """
    Convert language and metadata information to their rmd representation
    :param language:
    :param metadata:
    :return:
    """
    options = language.lower()
    metadata = copy(metadata)
    if 'name' in metadata:
        options += ' ' + metadata['name'] + ','
        del metadata['name']
    for jupyter_option, rmd_option, rev in _BOOLEAN_OPTIONS_DICTIONARY:
        if jupyter_option in metadata:
            options += ' {}={},'.format(
                rmd_option, _r_logical_values(metadata[jupyter_option] != rev))
            del metadata[jupyter_option]
    for opt_name in metadata:
        opt_value = metadata[opt_name]
        opt_name = opt_name.strip()
        if opt_name in _IGNORE_METADATA:
            continue
        elif opt_name == 'active':
            options += ' {}="{}",'.format(opt_name, str(opt_value))
        elif isinstance(opt_value, bool):
            options += ' {}={},'.format(
                opt_name, 'TRUE' if opt_value else 'FALSE')
        elif isinstance(opt_value, list):
            options += ' {}={},'.format(
                opt_name, 'c({})'.format(
                    ', '.join(['"{}"'.format(str(v)) for v in opt_value])))
        else:
            options += ' {}={},'.format(opt_name, str(opt_value))
    return options.strip(',').strip()


def update_md_from_rmd_options(name, value, metadata):
    """
    Update metadata using the _BOOLEAN_OPTIONS_DICTIONARY mapping
    :param name: option name
    :param value: option value
    :param metadata:
    :return:
    """
    for jupyter_option, rmd_option, rev in _BOOLEAN_OPTIONS_DICTIONARY:
        if name == rmd_option:
            try:
                metadata[jupyter_option] = _py_logical_values(value) != rev
                return True
            except RLogicalValueError:
                pass
    return False


class ParsingContext():
    """
    Class for determining where to split rmd options
    """
    parenthesis_count = 0
    curly_bracket_count = 0
    square_bracket_count = 0
    in_single_quote = False
    in_double_quote = False

    def __init__(self, line):
        self.line = line

    def in_global_expression(self):
        """Currently inside an expression"""
        return (self.parenthesis_count == 0 and self.curly_bracket_count == 0
                and self.square_bracket_count == 0
                and not self.in_single_quote and not self.in_double_quote)

    def count_special_chars(self, char, prev_char):
        """Update parenthesis counters"""
        if char == '(':
            self.parenthesis_count += 1
        elif char == ')':
            self.parenthesis_count -= 1
            if self.parenthesis_count < 0:
                raise RMarkdownOptionParsingError(
                    'Option line "{}" has too many '
                    'closing parentheses'.format(self.line))
        elif char == '{':
            self.curly_bracket_count += 1
        elif char == '}':
            self.curly_bracket_count -= 1
            if self.curly_bracket_count < 0:
                raise RMarkdownOptionParsingError(
                    'Option line "{}" has too many '
                    'closing curly brackets'.format(self.line))
        elif char == '[':
            self.square_bracket_count += 1
        elif char == ']':
            self.square_bracket_count -= 1
            if self.square_bracket_count < 0:
                raise RMarkdownOptionParsingError(
                    'Option line "{}" has too many '
                    'closing square brackets'.format(self.line))
        elif char == "'" and prev_char != '\\':
            self.in_single_quote = not self.in_single_quote
        elif char == '"' and prev_char != '\\':
            self.in_double_quote = not self.in_double_quote


def parse_rmd_options(line):
    """
    Given a R markdown option line, returns a list of pairs name,value
    :param line:
    :return:
    """
    parsing_context = ParsingContext(line)

    result = []
    prev_char = ''

    name = ''
    value = ''

    for char in ',' + line + ',':
        if parsing_context.in_global_expression():
            if char == ',':
                if name != '' or value != '':
                    if result and name == '':
                        raise RMarkdownOptionParsingError(
                            'Option line "{}" has no name for '
                            'option value {}'.format(line, value))
                    result.append((name.strip(), value.strip()))
                    name = ''
                    value = ''
            elif char == '=':
                if name == '':
                    name = value
                    value = ''
                else:
                    value += char
            else:
                parsing_context.count_special_chars(char, prev_char)
                value += char
        else:
            parsing_context.count_special_chars(char, prev_char)
            value += char
        prev_char = char

    if not parsing_context.in_global_expression():
        raise RMarkdownOptionParsingError(
            'Option line "{}" is not properly terminated'.format(line))

    return result


def rmd_options_to_metadata(options):
    """
    Parse rmd options and return a metadata dictionary
    :param options:
    :return:
    """
    options = re.split(r'\s|,', options, 1)
    if len(options) == 1:
        language = options[0]
        chunk_options = []
    else:
        language, others = options
        language = language.rstrip(' ,')
        others = others.lstrip(' ,')
        chunk_options = parse_rmd_options(others)

    language = 'R' if language == 'r' else language
    metadata = {}
    for i, opt in enumerate(chunk_options):
        name, value = opt
        if i == 0 and name == '':
            metadata['name'] = value
            continue
        else:
            if update_md_from_rmd_options(name, value, metadata):
                continue
            if name == 'active':
                metadata[name] = value.replace('"', '').replace("'", '')
                continue
            try:
                metadata[name] = _py_logical_values(value)
                continue
            except RLogicalValueError:
                metadata[name] = value

    for name in metadata:
        value = metadata[name]
        if not isinstance(value, str):
            continue
        if value.startswith('"') or value.startswith("'"):
            continue
        if value.startswith('c(') and value.endswith(')'):
            value = '[' + value[2:-1] + ']'
        elif value.startswith('list(') and value.endswith(')'):
            value = '[' + value[5:-1] + ']'
        try:
            metadata[name] = ast.literal_eval(value)
        except (SyntaxError, ValueError):
            continue

    if 'active' in metadata and 'eval' in metadata:
        del metadata['eval']

    return language, metadata


def json_options_to_metadata(options):
    """
    Read metadata from its json representation
    :param options:
    :return:
    """
    try:
        return json.loads(options)
    except ValueError:
        return {}


def metadata_to_json_options(metadata):
    """
    Represent metadata as json text
    :param metadata:
    :return:
    """
    return json.dumps({k: metadata[k] for k in metadata
                       if k not in _IGNORE_METADATA})



# coding: utf-8
import sys
import pytest
import nbrmd
from .utils import list_all_notebooks


@pytest.mark.parametrize('nb_file', list_all_notebooks('.ipynb') +
                         list_all_notebooks('.Rmd'))
def test_notebook_contents_is_unicode(nb_file):
    nb = nbrmd.readf(nb_file)

    for cell in nb.cells:
        if sys.version_info < (3, 0):
            assert cell.source == '' or isinstance(cell.source, unicode)
        else:
            assert isinstance(cell.source, str)


def test_write_non_ascii(tmpdir):
    nb = nbrmd.reads(u'Non-ascii contênt')
    nbrmd.writef(nb, str(tmpdir.join('notebook.Rmd')))
    nbrmd.writef(nb, str(tmpdir.join('notebook.ipynb')))



"""R markdown notebook format for Jupyter

Use this module to read or write Jupyter notebooks as R Markdown documents
(methods 'read', 'reads', 'write', 'writes')

Use the RmdFileContentsManager to open Rmd and Jupyter notebooks in Jupyter

Use the 'nbrmd' conversion script to convert Jupyter notebooks from/to
R Markdown notebooks.
"""

from .nbrmd import readf, writef, writes, reads, NOTEBOOK_EXTENSIONS, readme

try:
    from .rmarkdownexporter import RMarkdownExporter
except ImportError as err:
    RMarkdownExporter = str(err)

try:
    from .contentsmanager import RmdFileContentsManager
except ImportError as err:
    RmdFileContentsManager = str(err)

__all__ = ['readf', 'writef', 'writes', 'reads', 'NOTEBOOK_EXTENSIONS',
           'readme', 'RMarkdownExporter', 'RmdFileContentsManager']



from setuptools import setup, find_packages
from nbrmd.nbrmd import readme

setup(
    name='nbrmd',
    version='0.4.4',
    author='Marc Wouts',
    author_email='marc.wouts@gmail.com',
    description='Jupyter from/to R markdown notebooks',
    long_description=readme(),
    long_description_content_type='text/markdown',
    url='https://github.com/mwouts/nbrmd',
    packages=find_packages(),
    entry_points={'console_scripts': ['nbrmd = nbrmd.cli:main'],
                  'nbconvert.exporters':
                      ['rmarkdown = nbrmd:RMarkdownExporter']},
    tests_require=['pytest', 'testfixtures'],
    install_requires=['nbformat>=4.0.0', 'mock', 'pyyaml'],
    license='MIT',
    classifiers=('Development Status :: 4 - Beta',
                 'Environment :: Console',
                 'Framework :: Jupyter',
                 'Intended Audience :: Science/Research',
                 'Programming Language :: Python',
                 'Topic :: Text Processing :: Markup',
                 'Programming Language :: Python',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6')
)



"""ContentsManager that allows to open Rmd, py, R and ipynb files as notebooks
"""
import os
import nbformat
import mock

try:
    import notebook.transutils  # noqa
except ImportError:
    pass
from notebook.services.contents.filemanager import FileContentsManager
from traitlets import Unicode
from traitlets.config import Configurable
import nbrmd

from . import combine


def _nbrmd_writes(ext):
    def _writes(nb, version=nbformat.NO_CONVERT, **kwargs):
        return nbrmd.writes(nb, version=version, ext=ext, **kwargs)

    return _writes


def _nbrmd_reads(ext):
    def _reads(text, as_version, **kwargs):
        return nbrmd.reads(text, as_version, ext=ext, **kwargs)

    return _reads


def check_formats(formats):
    """
    Parse, validate and return notebooks extensions
    :param formats: a list of notebook extensions, or a comma separated string
    :return: list of extensions
    """
    if not isinstance(formats, list):
        formats = formats.split(',')

    formats = [fmt if fmt.startswith('.') else '.' + fmt
               for fmt in formats if fmt != '']

    allowed = nbrmd.NOTEBOOK_EXTENSIONS
    if not isinstance(formats, list) or not set(formats).issubset(allowed):
        raise TypeError("Notebook metadata 'nbrmd_formats' "
                        "should be subset of {}, but was {}"
                        "".format(str(allowed), str(formats)))
    return formats


class RmdFileContentsManager(FileContentsManager, Configurable):
    """
    A FileContentsManager Class that reads and stores notebooks to classical
    Jupyter notebooks (.ipynb), R Markdown notebooks (.Rmd),
    R scripts (.R) and python scripts (.py)
    """

    nb_extensions = [ext for ext in nbrmd.NOTEBOOK_EXTENSIONS if
                     ext != '.ipynb']

    def all_nb_extensions(self):
        """
        Notebook extensions, including ipynb
        :return:
        """
        return ['.ipynb'] + self.nb_extensions

    default_nbrmd_formats = Unicode(
        u'ipynb',
        help='Save notebooks to these file extensions. '
             'Can be any of ipynb,Rmd,py,R, comma separated',
        config=True)

    def _read_notebook(self, os_path, as_version=4,
                       load_alternative_format=True):
        """Read a notebook from an os path."""
        file, ext = os.path.splitext(os_path)
        if ext in self.nb_extensions:
            with mock.patch('nbformat.reads', _nbrmd_reads(ext)):
                nb = super(RmdFileContentsManager, self) \
                    ._read_notebook(os_path, as_version)
        else:
            nb = super(RmdFileContentsManager, self) \
                ._read_notebook(os_path, as_version)

        if not load_alternative_format:
            return nb

        # Notebook formats: default, notebook metadata, or current extension
        nbrmd_formats = (nb.metadata.get('nbrmd_formats') or
                         self.default_nbrmd_formats)

        nbrmd_formats = check_formats(nbrmd_formats)

        if ext not in nbrmd_formats:
            nbrmd_formats.append(ext)

        nbrmd_formats = check_formats(nbrmd_formats)

        # Source format is current ext, or is first non .ipynb format
        # that is found on disk
        source_format = None
        if ext != '.ipynb':
            source_format = ext
        else:
            for fmt in nbrmd_formats:
                if fmt != '.ipynb' and os.path.isfile(file + fmt):
                    source_format = fmt
                    break

        nb_outputs = None
        if source_format is not None and ext != source_format:
            self.log.info('Reading SOURCE from {}'
                          .format(os.path.basename(file + source_format)))
            self.log.info('Reading OUTPUTS from {}'
                          .format(os.path.basename(os_path)))
            nb_outputs = nb
            nb = self._read_notebook(file + source_format,
                                     as_version=as_version,
                                     load_alternative_format=False)
        elif ext != '.ipynb' and '.ipynb' in nbrmd_formats \
                and os.path.isfile(file + '.ipynb'):
            self.log.info('Reading SOURCE from {}'
                          .format(os.path.basename(os_path)))
            self.log.info('Reading OUTPUTS from {}'
                          .format(os.path.basename(file + '.ipynb')))
            nb_outputs = self._read_notebook(file + '.ipynb',
                                             as_version=as_version,
                                             load_alternative_format=False)

        if nb_outputs is not None:
            combine.combine_inputs_with_outputs(nb, nb_outputs)
            if self.notary.check_signature(nb_outputs):
                self.notary.sign(nb)

        return nb

    def _save_notebook(self, os_path, nb):
        """Save a notebook to an os_path."""
        os_file, org_ext = os.path.splitext(os_path)

        formats = (nb.get('metadata', {}).get('nbrmd_formats') or
                   self.default_nbrmd_formats)

        formats = check_formats(formats)

        if org_ext not in formats:
            formats.append(org_ext)

        formats = check_formats(formats)

        for ext in formats:
            os_path_ext = os_file + ext
            self.log.info("Saving %s", os.path.basename(os_path_ext))
            if ext in self.nb_extensions:
                with mock.patch('nbformat.writes', _nbrmd_writes(ext)):
                    super(RmdFileContentsManager, self) \
                        ._save_notebook(os_path_ext, nb)
            else:
                super(RmdFileContentsManager, self) \
                    ._save_notebook(os_path_ext, nb)

    def get(self, path, content=True, type=None, format=None):
        """ Takes a path for an entity and returns its model"""
        path = path.strip('/')

        if self.exists(path) and \
                (type == 'notebook' or
                 (type is None and
                  any([path.endswith(ext)
                       for ext in self.all_nb_extensions()]))):
            return self._notebook_model(path, content=content)

        return super(RmdFileContentsManager, self) \
            .get(path, content, type, format)

    def trust_notebook(self, path):
        """Trust the current notebook"""
        file, _ = os.path.splitext(path)
        super(RmdFileContentsManager, self).trust_notebook(file + '.ipynb')

    def rename_file(self, old_path, new_path):
        """Rename the current notebook, as well as its
         alternative representations"""
        old_file, org_ext = os.path.splitext(old_path)
        new_file, new_ext = os.path.splitext(new_path)
        if org_ext in self.all_nb_extensions() and org_ext == new_ext:
            for ext in self.all_nb_extensions():
                if self.file_exists(old_file + ext):
                    super(RmdFileContentsManager, self) \
                        .rename_file(old_file + ext, new_file + ext)
        else:
            super(RmdFileContentsManager, self).rename_file(old_path, new_path)



"""Cell conversion methods.

Convert/read notebook cells to/from python, R and Rmd cells
"""

import re
from nbformat.v4.nbbase import new_code_cell, new_raw_cell, new_markdown_cell
from .languages import cell_language, is_code
from .cell_metadata import metadata_to_rmd_options, rmd_options_to_metadata, \
    json_options_to_metadata, metadata_to_json_options


def code_to_rmd(source, metadata, default_language):
    """
    Represent a code cell with given source and metadata as a rmd cell
    :param source:
    :param metadata:
    :param default_language:
    :return:
    """
    lines = []
    language = cell_language(source) or default_language
    if 'Rmd' not in re.split('\\.|,', metadata.get('active', 'Rmd')):
        metadata['eval'] = False
    options = metadata_to_rmd_options(language, metadata)
    lines.append(u'```{{{}}}'.format(options))
    lines.extend(source)
    lines.append(u'```')
    return lines


def code_to_text(self,
                 source,
                 metadata,
                 default_language,
                 next_cell_is_code):
    """
    Represent a code cell with given source and metadata as text
    :param self:
    :param source:
    :param metadata:
    :param default_language:
    :param next_cell_is_code:
    :return:
    """
    if self.ext == '.Rmd':
        return code_to_rmd(source, metadata, default_language)
    else:
        lines = []
        language = cell_language(source) or default_language
        if language == default_language:
            if self.ext == '.R':
                options = metadata_to_rmd_options(language, metadata)[2:]
                if options != '':
                    lines.append('#+ ' + options)
            else:
                options = metadata_to_json_options(metadata)
                if options != '{}':
                    lines.append('# + ' + options)
            lines.extend(source)
        else:
            lines.extend(self.markdown_escape(
                code_to_rmd(source, metadata, default_language)))

        # Two blank lines before next code cell
        if next_cell_is_code:
            lines.append('')

        return lines


def cell_to_text(self,
                 cell,
                 next_cell=None,
                 default_language='python'):
    """
    Represent a notebook cell as a text cell, in either
    py, R or rmd format
    :param self: TextNotebookWriter object
    :param cell: current cell
    :param next_cell: next cell
    :param default_language: default language for the current notebook
    :return:
    """
    source = cell.get('source').splitlines()
    metadata = cell.get('metadata', {})
    skipline = True
    if 'noskipline' in metadata:
        skipline = not metadata['noskipline']
        del metadata['noskipline']

    lines = []
    if is_code(cell):
        lines.extend(code_to_text(self, source, metadata, default_language,
                                  next_cell and is_code(next_cell)))
    else:
        if source == []:
            source = ['']
        lines.extend(self.markdown_escape(source))

        # Two blank lines between consecutive markdown cells in Rmd
        if self.ext == '.Rmd' and next_cell \
                and next_cell.cell_type == 'markdown':
            lines.append('')

    if skipline and next_cell:
        lines.append('')

    return lines


_START_CODE_RMD = re.compile(r"^```\{(.*)\}\s*$")
_END_CODE_MD = re.compile(r"^```\s*$")
_CODE_OPTION_RPY = re.compile(r"^(#|# )\+(.*)$")
_BLANK_LINE = re.compile(r"^\s*$")


def start_code_rmd(line):
    """
    Line indicates that a code cell starts, in a rmd file
    :param line:
    :return:
    """
    return _START_CODE_RMD.match(line)


def start_code_rpy(line):
    """
    A code cell starts here, in a py or R file
    :param line:
    :return:
    """
    return _CODE_OPTION_RPY.match(line)


def next_uncommented_is_code(lines):
    """
    Next non-commented line is code
    :param lines:
    :return:
    """
    for line in lines:
        if line.startswith('#'):
            continue
        return not _BLANK_LINE.match(line)

    return False


def text_to_cell(self, lines):
    """
    Parse text to a cell
    :param self:
    :param lines:
    :return: cell, cursor
    """
    if self.start_code(lines[0]):
        return self.code_to_cell(lines, parse_opt=True)
    elif self.prefix != '' and not lines[0].startswith(self.prefix):
        return self.code_to_cell(lines, parse_opt=False)
    elif self.ext == '.py' and next_uncommented_is_code(lines):
        return self.code_to_cell(lines, parse_opt=False)

    return self.markdown_to_cell(lines)


def parse_code_options(line, ext):
    """
    Parse code options on the given line
    :param line:
    :param ext:
    :return:
    """
    if ext == '.Rmd':
        return rmd_options_to_metadata(_START_CODE_RMD.findall(line)[0])
    elif ext == '.R':
        return rmd_options_to_metadata(_CODE_OPTION_RPY.match(line).group(2))

    return 'python', json_options_to_metadata(
        _CODE_OPTION_RPY.match(line).group(2))


def next_code_is_indented(lines):
    """
    Is next code line indented?
    :param lines:
    :return:
    """
    for line in lines:
        if line.startswith('#'):
            continue
        if _BLANK_LINE.match(line):
            continue
        return line.startswith(' ')

    return False


def no_code_before_next_blank_line(lines):
    """
    Do we find code before next blank line?
    :param lines:
    :return:
    """
    for line in lines:
        if line.startswith('#'):
            continue
        return _BLANK_LINE.match(line)

    return True


def code_or_raw_cell(source, metadata):
    if 'ipynb' not in re.split('\\.|,', metadata.get('active', 'ipynb')):
        return new_raw_cell(source=source, metadata=metadata)
    return new_code_cell(source=source, metadata=metadata)


def code_to_cell(self, lines, parse_opt):
    """
    Parse code to a notebook cell
    :param self:
    :param lines:
    :param parse_opt:
    :return: cell, cursor
    """
    # Parse options
    if parse_opt:
        language, metadata = parse_code_options(lines[0], self.ext)
        if self.ext == '.Rmd':
            metadata['language'] = language
    else:
        metadata = {}

    # Find end of cell and return
    if self.ext == '.Rmd':
        for pos, line in enumerate(lines):
            if pos > 0 and _END_CODE_MD.match(line):
                next_line_blank = pos + 1 == len(lines) or \
                                  _BLANK_LINE.match(lines[pos + 1])
                if next_line_blank and pos + 2 != len(lines):
                    return code_or_raw_cell(
                        source='\n'.join(lines[1:pos]), metadata=metadata), \
                           pos + 2
                cell = code_or_raw_cell(
                    source='\n'.join(lines[1:pos]),
                    metadata=metadata)
                cell.metadata['noskipline'] = True
                return cell, pos + 1
    else:
        prev_blank = False
        triple = None
        for pos, line in enumerate(lines):
            if parse_opt and pos == 0:
                continue

            if self.ext == '.R' and line.startswith(self.prefix):
                if prev_blank:
                    return code_or_raw_cell(
                        source='\n'.join(lines[parse_opt:(pos - 1)]),
                        metadata=metadata), pos
                cell = code_or_raw_cell(
                    source='\n'.join(lines[parse_opt:pos]),
                    metadata=metadata)
                cell.metadata['noskipline'] = True
                return cell, pos

            if self.ext == '.py':
                single = None
                for i, char in enumerate(line):
                    if char not in ['"', "'"]:
                        continue

                    if single == char:
                        single = None
                        continue
                    if single is not None:
                        continue
                    if triple == char:
                        if line[i - 2:i + 1] == 3 * char:
                            triple = None
                            continue
                    if triple is not None:
                        continue
                    if line[i - 2:i + 1] == 3 * char:
                        triple = char

                if triple:
                    continue

            if prev_blank:
                if _BLANK_LINE.match(line):
                    # Two blank lines => end of cell
                    # (unless next code is indented)
                    # Two blank lines at the end == empty code cell

                    if self.ext == '.py':
                        if next_code_is_indented(lines[pos:]):
                            continue

                    return code_or_raw_cell(
                        source='\n'.join(lines[parse_opt:(pos - 1)]),
                        metadata=metadata), min(pos + 1,
                                                len(lines) - 1)

                # are all the lines from here to next blank
                # escaped with the prefix?
                if self.ext == '.py':
                    if no_code_before_next_blank_line(lines[pos:]):
                        return code_or_raw_cell(
                            source='\n'.join(lines[parse_opt:(pos - 1)]),
                            metadata=metadata), pos

            prev_blank = _BLANK_LINE.match(line)

    # Unterminated cell?
    return code_or_raw_cell(
        source='\n'.join(lines[parse_opt:]),
        metadata=metadata), len(lines)


def markdown_to_cell(self, lines):
    """
    Parse text and return a markdown cell
    :param self:
    :param lines:
    :return: cell, cursor
    """
    markdown = []
    for pos, line in enumerate(lines):
        # Markdown stops with the end of comments
        if line.startswith(self.prefix):
            markdown.append(self.markdown_unescape(line))
        elif _BLANK_LINE.match(line):
            return new_markdown_cell(source='\n'.join(markdown)), pos + 1
        else:
            cell = new_markdown_cell(source='\n'.join(markdown))
            cell.metadata['noskipline'] = True
            return cell, pos

    # still here => unterminated markdown
    return new_markdown_cell(source='\n'.join(markdown)), len(lines)


def markdown_to_cell_rmd(lines):
    """
    Parse text, in case of a rmd file, and return a markdown cell
    :param lines:
    :return: cell, cursor
    """
    prev_blank = False
    for pos, line in enumerate(lines):
        if start_code_rmd(line):
            if prev_blank and pos > 1:
                return new_markdown_cell(
                    source='\n'.join(lines[:(pos - 1)])), pos
            cell = new_markdown_cell(
                source='\n'.join(lines[:pos]))
            cell.metadata['noskipline'] = True
            return cell, pos

        if _BLANK_LINE.match(line) and prev_blank:
            return new_markdown_cell(
                source='\n'.join(lines[:(pos - 1)])), pos + 1
        prev_blank = _BLANK_LINE.match(line)

    # Unterminated cell?
    return new_markdown_cell(source='\n'.join(lines)), len(lines)



"""Read and write notebooks as RStudio notebook files, with .Rmd extension.

Raw and markdown cells are converted to markdown, while code cells are
converted to code chunks. The transformation is reversible and all inputs
are preserved (not outputs, though).

Authors:

* Marc Wouts
"""

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

import os
import io
from copy import copy
from nbformat.v4.rwbase import NotebookReader, NotebookWriter
from nbformat.v4.nbbase import new_notebook
import nbformat

from .header import header_to_metadata_and_cell, metadata_and_cell_to_header, \
    encoding_and_executable
from .languages import get_default_language, find_main_language
from .cells import start_code_rmd, start_code_rpy, cell_to_text, text_to_cell
from .cells import markdown_to_cell_rmd, markdown_to_cell, code_to_cell

# -----------------------------------------------------------------------------
# Code
# -----------------------------------------------------------------------------


NOTEBOOK_EXTENSIONS = ['.ipynb', '.Rmd', '.py', '.R']


def markdown_comment(ext):
    """Markdown escape for given notebook extension"""
    return '' if ext == '.Rmd' else "#'" if ext == '.R' else "#"


class TextNotebookReader(NotebookReader):
    """Text notebook reader"""

    def __init__(self, ext):
        self.ext = ext
        self.prefix = markdown_comment(ext)
        self.start_code = start_code_rmd if ext == '.Rmd' else start_code_rpy
        if ext == '.Rmd':
            self.markdown_to_cell = markdown_to_cell_rmd

    header_to_metadata_and_cell = header_to_metadata_and_cell
    text_to_cell = text_to_cell
    code_to_cell = code_to_cell
    markdown_to_cell = markdown_to_cell

    def markdown_unescape(self, line):
        """Remove markdown escape, if any"""
        if self.prefix == '':
            return line
        line = line[len(self.prefix):]
        if line.startswith(' '):
            line = line[1:]
        return line

    def reads(self, s, **kwargs):
        """
        Read a notebook from a given string
        :param s:
        :param kwargs:
        :return:
        """
        return self.to_notebook(s, **kwargs)

    def to_notebook(self, text, **kwargs):
        """
        Read a notebook from its text representation
        :param text:
        :param kwargs:
        :return:
        """
        lines = text.splitlines()

        cells = []
        metadata, header_cell, pos = \
            self.header_to_metadata_and_cell(lines)

        if header_cell:
            cells.append(header_cell)

        if pos > 0:
            lines = lines[pos:]

        while lines:
            cell, pos = self.text_to_cell(lines)
            if cell is None:
                break
            cells.append(cell)
            if pos <= 0:
                raise Exception('Blocked at lines ' + '\n'.join(lines[:6]))
            lines = lines[pos:]

        if self.ext == '.Rmd':
            find_main_language(metadata, cells)

        nb = new_notebook(cells=cells, metadata=metadata)
        return nb


class TextNotebookWriter(NotebookWriter):
    """Write notebook to their text representations"""

    def __init__(self, ext='.Rmd'):
        self.ext = ext
        self.prefix = markdown_comment(ext)

    def markdown_escape(self, lines):
        """
        Escape markdown text
        :param lines:
        :return:
        """
        if self.prefix == '':
            return lines
        return [self.prefix if line == '' else self.prefix + ' ' + line
                for line in lines]

    encoding_and_executable = encoding_and_executable
    metadata_and_cell_to_header = metadata_and_cell_to_header
    cell_to_text = cell_to_text

    def writes(self, nb, **kwargs):
        """Write the text representation of a notebook to a string"""
        nb = copy(nb)
        if self.ext == '.py':
            default_language = 'python'
        elif self.ext == '.R':
            default_language = 'R'
        else:
            default_language = get_default_language(nb)

        lines = self.encoding_and_executable(nb)
        lines.extend(self.metadata_and_cell_to_header(nb))

        for i in range(len(nb.cells)):
            cell = nb.cells[i]
            next_cell = nb.cells[i + 1] if i + 1 < len(nb.cells) else None
            lines.extend(self.cell_to_text(cell, next_cell,
                                           default_language=default_language))

        return '\n'.join(lines + [''])


_NOTEBOOK_READERS = {ext: TextNotebookReader(ext)
                     for ext in NOTEBOOK_EXTENSIONS if ext != '.ipynb'}
_NOTEBOOK_WRITERS = {ext: TextNotebookWriter(ext)
                     for ext in NOTEBOOK_EXTENSIONS if ext != '.ipynb'}


def reads(text, as_version=4, ext='.Rmd', **kwargs):
    """Read a notebook from a string"""
    if ext == '.ipynb':
        return nbformat.reads(text, as_version, **kwargs)

    return _NOTEBOOK_READERS[ext].reads(text, **kwargs)


def read(fp, as_version=4, ext='.Rmd', **kwargs):
    """Read a notebook from a file"""
    if ext == '.ipynb':
        return nbformat.read(fp, as_version, **kwargs)

    return _NOTEBOOK_READERS[ext].read(fp, **kwargs)


def writes(nb, version=nbformat.NO_CONVERT, ext='.Rmd', **kwargs):
    """Write a notebook to a string"""
    if ext == '.ipynb':
        return nbformat.writes(nb, version, **kwargs)

    return _NOTEBOOK_WRITERS[ext].writes(nb)


def write(np, fp, version=nbformat.NO_CONVERT, ext='.Rmd', **kwargs):
    """Write a notebook to a file"""
    if ext == '.ipynb':
        return nbformat.write(np, fp, version, **kwargs)

    return _NOTEBOOK_WRITERS[ext].write(np, fp)


def readf(nb_file):
    """Read a notebook from the file with given name"""
    _, ext = os.path.splitext(nb_file)
    if ext not in NOTEBOOK_EXTENSIONS:
        raise TypeError(
            'File {} is not a notebook. '
            'Expected extensions are {}'.format(nb_file,
                                                NOTEBOOK_EXTENSIONS))
    with io.open(nb_file, encoding='utf-8') as fp:
        return read(fp, as_version=4, ext=ext)


def writef(nb, nb_file):
    """Write a notebook to the file with given name"""
    _, ext = os.path.splitext(nb_file)
    if ext not in NOTEBOOK_EXTENSIONS:
        raise TypeError(
            'File {} is not a notebook. '
            'Expected extensions are {}'.format(nb_file,
                                                NOTEBOOK_EXTENSIONS))
    with io.open(nb_file, 'w', encoding='utf-8') as fp:
        write(nb, fp, version=nbformat.NO_CONVERT, ext=ext)


def readme():
    """Contents of README.md"""
    readme_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                               '..', 'README.md')
    with open(readme_path) as readme_file:
        return readme_file.read()



import os
import sys
import pytest
from nbrmd import RmdFileContentsManager, readf
from .utils import list_all_notebooks, remove_outputs


@pytest.mark.skipif(isinstance(RmdFileContentsManager, str),
                    reason=RmdFileContentsManager)
def test_create_contentsmanager():
    RmdFileContentsManager()


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.skipif(isinstance(RmdFileContentsManager, str),
                    reason=RmdFileContentsManager)
@pytest.mark.parametrize('nb_file', list_all_notebooks('.ipynb'))
def test_load_save_rename(nb_file, tmpdir):
    tmp_ipynb = 'notebook.ipynb'
    tmp_rmd = 'notebook.Rmd'

    cm = RmdFileContentsManager()
    cm.root_dir = str(tmpdir)

    # open ipynb, save Rmd, reopen
    nb = readf(nb_file)
    cm.save(model=dict(type='notebook', content=nb), path=tmp_rmd)
    nb_rmd = cm.get(tmp_rmd)
    assert remove_outputs(nb) == remove_outputs(nb_rmd['content'])

    # save ipynb
    cm.save(model=dict(type='notebook', content=nb), path=tmp_ipynb)

    # rename ipynb
    cm.rename(tmp_ipynb, 'new.ipynb')
    assert not os.path.isfile(str(tmpdir.join(tmp_ipynb)))
    assert not os.path.isfile(str(tmpdir.join(tmp_rmd)))

    assert os.path.isfile(str(tmpdir.join('new.ipynb')))
    assert os.path.isfile(str(tmpdir.join('new.Rmd')))



import sys
import pytest
from nbrmd.cell_metadata import rmd_options_to_metadata, \
    metadata_to_rmd_options, parse_rmd_options, RMarkdownOptionParsingError

SAMPLES = [('r', ('R', {})),
           ('r plot_1, dpi=72, fig.path="fig_path/"',
            ('R', {'name': 'plot_1', 'dpi': 72, 'fig.path': '"fig_path/"'})),
           ("r plot_1, bool=TRUE, fig.path='fig_path/'",
            ('R', {'name': 'plot_1', 'bool': True,
                   'fig.path': "'fig_path/'"})),
           ('r echo=FALSE',
            ('R', {'hide_input': True})),
           ('r plot_1, echo=TRUE',
            ('R', {'name': 'plot_1', 'hide_input': False})),
           ('python echo=if a==5 then TRUE else FALSE',
            ('python', {'echo': 'if a==5 then TRUE else FALSE'})),
           ('python noname, tags=c("a", "b", "c"), echo={sum(a+c(1,2))>1}',
            ('python', {'name': 'noname', 'tags': ['a', 'b', 'c'],
                        'echo': '{sum(a+c(1,2))>1}'})),
           ('python active="ipynb,py"',
            ('python', {'active': 'ipynb,py'})),
           ('python include=FALSE, active="Rmd"',
            ('python', {'active': 'Rmd', 'hide_output': True})),
           ('r chunk_name, include=FALSE, active="Rmd"',
            ('R',
             {'name': 'chunk_name', 'active': 'Rmd', 'hide_output': True}))
           ]


@pytest.mark.parametrize('options,language_and_metadata', SAMPLES)
def test_parse_rmd_options(options, language_and_metadata):
    assert rmd_options_to_metadata(options) == language_and_metadata


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.parametrize('options,language_and_metadata', SAMPLES)
def test_build_options(options, language_and_metadata):
    assert metadata_to_rmd_options(*language_and_metadata) == options


@pytest.mark.parametrize('options,language_and_metadata', SAMPLES)
def test_build_options_random_order(options, language_and_metadata):
    # Older python has no respect for order...
    # assert to_chunk_options(metadata) == options

    def split_and_strip(opt):
        set([o.strip() for o in opt.split(',')])

    assert (split_and_strip(metadata_to_rmd_options(*language_and_metadata)) ==
            split_and_strip(options))


@pytest.mark.parametrize('options', ['a={)', 'name, name2',
                                     'a=}', 'b=]', 'c=['])
def test_parsing_error(options):
    with pytest.raises(RMarkdownOptionParsingError):
        parse_rmd_options(options)
